/*
   InformixInt.ec ---

   This TCL-extension is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Written by Joachim Siebold <joachim.siebold@cityweb.de>
   based on DBD::Informix (Copyright J. Leffler and others)

   $Header: /home/cvs/informixtcl/InformixInt.ec,v 1.5 2000/03/22 12:58:53 achim Exp $
   $Log: InformixInt.ec,v $
   Revision 1.5  2000/03/22 12:58:53  achim
   *** empty log message ***

   Revision 1.4  2000/03/22 12:54:46  achim
   *** empty log message ***

   Revision 1.3  2000/03/19 14:11:40  achim
   Clean trailing cr/lf

   Revision 1.2  2000/03/18 18:23:57  achim
   Introduced CVS Keywords

*/         

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include <tcl.h>
#include "InformixTcl.h" 

#define L_CURLY '{'
#define R_CURLY '}'   

#define R_BINARY 0
#define R_CHAR   1
#define R_DOUBLE 2
#define R_LONG   3
#define R_INT    4

/* Internal implementation of BEGIN WORK */
/* Assumes correct connection is already set */

int 
InformixIntBegin(logonVars)
    LogonInfo       *logonVars;     /* logon handle */
{
    static char     funct[]= "InformixIntBegin";
    int             rc = 1;

    EXEC SQL BEGIN WORK;
    if (sqlca.sqlcode < 0)
        rc = 0;
    else
    {
        InformixApiDebug(3, "%s: BEGIN WORK\n", funct);
        logonVars->is_txactive = True;
    }
    return rc;
}         

/* Internal implementation of COMMIT WORK */
/* Assumes correct connection is already set */

int 
InformixIntCommit(logonVars)
    LogonInfo       *logonVars;     /* logon handle */
{
    static char     funct[]= "InformixIntCommit";
    int             rc = 1;

    EXEC SQL COMMIT WORK;
    if (sqlca.sqlcode < 0)
        rc = 0;
    else
    {
        InformixApiDebug(3, "%s: COMMIT WORK\n", funct);
        logonVars->is_txactive = False;
    }
    return rc;
}         

/* Internal implementation of Rollback WORK */
/* Assumes correct connection is already set */

int 
InformixIntRollback(logonVars)
    LogonInfo       *logonVars;     /* logon handle */
{
    static char     funct[]= "InformixIntRollback";
    int             rc = 1;

    EXEC SQL ROLLBACK WORK;
    if (sqlca.sqlcode < 0)
        rc = 0;
    else
    {
        InformixApiDebug(3, "%s: ROLLBACK WORK\n", funct);
        logonVars->is_txactive = False;
    }
    return rc;
}         

/*
 *--------------------------------------------------------------
 *
 * InformixIntPrepare
 *
 * Prepares a new Statement (Internal Representation)
 *
 * Results:
 * Returns 0 on error 1 on success
 *
 *--------------------------------------------------------------
*/
int
InformixIntPrepare(statementVars, cmd)
    StatementInfo   *statementVars;     /* Statement handle data */
    char            *cmd;               /* SQL Statement */
{
    int             rc=1;               /* Return Value */
    static char function[] = "InformixIntPrepare";  

    EXEC SQL BEGIN DECLARE SECTION;
    char           *statement = cmd;
    int             desc_count;
    char           *nm_stmnt;
    char           *nm_obind;
    char           *nm_cursor;
    EXEC SQL END DECLARE SECTION;  

    InformixApiEnter(function);

    nm_stmnt = statementVars->nm_stmnt;
    nm_obind = statementVars->nm_obind;
    nm_cursor = statementVars->nm_cursor;

    /* Record the number of input parameters in the statement */
    if(InformixIntPreparse(statementVars,statement)==0) 
    {
        InformixApiExit(function);
        return 0;
    }
    
    /* Allocate space for that many parameters */
    if (InformixIntSetBindNum(statementVars, statementVars->n_input_params) == 0)
    {
        InformixApiExit(function);
        return 0;
    }

    InformixApiDebug(4, "InformixIntPrepare -- <<%s>>\n", statement);
    EXEC SQL PREPARE :nm_stmnt FROM :statement;
    InformixApiSaveSqlca(statementVars->logonVars);

    if (sqlca.sqlcode < 0)
    {
        InformixApiSqlCode(statementVars->logonVars);
        InformixApiExit(function);
        return 0;
    }
    statementVars->st_state = Prepared;

    EXEC SQL ALLOCATE DESCRIPTOR :nm_obind WITH MAX 128;
    if (sqlca.sqlcode < 0)
    {
        InformixIntDelStatement(statementVars);
        InformixApiSqlCode(statementVars->logonVars);
        InformixApiExit(function);
        return 0;
    }
    statementVars->st_state = Allocated;    

    EXEC SQL DESCRIBE :nm_stmnt USING SQL DESCRIPTOR :nm_obind;
    if (sqlca.sqlcode < 0)
    {
        InformixIntDelStatement(statementVars);
        InformixApiSqlCode(statementVars->logonVars);
        InformixApiExit(function);
        return 0;
    }
    statementVars->st_state = Described;  
    statementVars->st_type= sqlca.sqlcode;
    if(statementVars->st_type == 0) statementVars->st_type= SQ_SELECT;

    EXEC SQL GET DESCRIPTOR :nm_obind :desc_count = COUNT;
    InformixApiSqlCode(statementVars->logonVars);
    if (sqlca.sqlcode < 0)
    {
        InformixIntDelStatement(statementVars);
        InformixApiSqlCode(statementVars->logonVars);
        InformixApiExit(function);
        return 0;
    }

    statementVars->n_columns = desc_count;  

    /**
    ** Only non-cursory statements need an output descriptor.
    ** Only cursory statements need a cursor declared for them.
    ** INSERT may need an input descriptor (which will appear to be the
    ** output descriptor, such being the wonders of Informix).
    */
    if (statementVars->st_type == SQ_SELECT)
        rc = InformixIntDeclare(statementVars);
    else if (statementVars->st_type == SQ_EXECPROC && desc_count > 0)
        rc = InformixIntDeclare(statementVars);
    else if (statementVars->st_type == SQ_INSERT && desc_count > 0)
    {
        InformixIntBlobs(statementVars);
        if (statementVars->n_blobs > 0)
        {
            /*
            ** Switch the nm_obind and nm_ibind names so that when
            ** INFORMIX_bindsv() is at work, it has an already populated
            ** SQL descriptor to work with, that already has the blobs
            ** set up correctly.
            */
            Name tmpname;
            strcpy(tmpname, statementVars->nm_ibind);
            strcpy(statementVars->nm_ibind, statementVars->nm_obind);
            strcpy(statementVars->nm_obind, tmpname);
            statementVars->n_bound = desc_count;
        }
        rc = 1;
    }
    else           
    {
        EXEC SQL DEALLOCATE DESCRIPTOR :nm_obind;
        statementVars->st_state = Prepared;
        rc = 1;
    }

    /* Get number of fields and space needed for field names      */

    InformixApiExit(function);
    return rc;
}   
/*
 *--------------------------------------------------------------
 *
 * InformixIntDelStatement
 *
 * Deletes Statement
 *
 * Results:
 *
 *--------------------------------------------------------------
*/
int
InformixIntDelStatement(statementVars)
    StatementInfo   *statementVars;     /* Statement handle data */
{
    static char function[] = "InformixApiPrepare";  

#if ESQLC_VERSION < 724
#define INFORMIX_RELEASE_BLOBS
#endif

    EXEC SQL BEGIN DECLARE SECTION;
    char           *name;
#ifdef INFORMIX_RELEASE_BLOBS
    int colno;
    int coltype;
    loc_t   blob;
#endif /* INFORMIX_RELEASE_BLOBS */
    EXEC SQL END DECLARE SECTION;
                                                 
    InformixApiDebug(3,function, "enter function");

    
    InformixIntSetConnection(statementVars->logonVars->nm_connection);

    switch (statementVars->st_state)      
    {
    case Finished:
        /*FALLTHROUGH*/

    case Opened:
        InformixApiDebug(3, function, " close cursor ");
        name = statementVars->nm_cursor;
        EXEC SQL CLOSE :name;
        /*FALLTHROUGH*/

    case Declared:
        InformixApiDebug(3, function, " free cursor ");
        name = statementVars->nm_cursor;
        EXEC SQL FREE :name;
        /*FALLTHROUGH*/

    case Described:
    case Allocated:
        name = statementVars->nm_obind;          
        /* Verified unfixed for ESQL/C 6.00.UE1 on Solaris 2.4 */
        /* Verified unfixed for ESQL/C 7.21.UC1 on Solaris 2.4 with Purify */
        /* Verified unfixed for ESQL/C 7.23.UC1 on Solaris 2.4 */
        /* Verified *fixed* in 7.24.UC1 on Solaris 2.5.1 (bad frees reported) */
        /* Verified as a bad fix on Windows 95/NT by Harald Ums (no version) */

#ifdef INFORMIX_RELEASE_BLOBS
        if (statementVars->n_blobs > 0)
        {
            for (colno = 1; colno <= statementVars->n_columns; colno++)
            {
                EXEC SQL GET DESCRIPTOR :name VALUE :colno :coltype = TYPE;
                if (coltype == SQLBYTES || coltype == SQLTEXT)
                {
                    EXEC SQL GET DESCRIPTOR :name VALUE :colno :blob = DATA;
                    if (blob.loc_loctype == LOCMEMORY && blob.loc_buffer != 0)
                        free(blob.loc_buffer);
                }
            }
        }
#endif /* INFORMIX_RELEASE_BLOBS */

        InformixApiDebug(3, function, " deallocate descriptor");
        EXEC SQL DEALLOCATE DESCRIPTOR :name;
        /*FALLTHROUGH*/

    case Prepared:
        InformixApiDebug(3,function, " free statement ");
        name = statementVars->nm_stmnt;
        EXEC SQL FREE :name;
        /*FALLTHROUGH*/

    case Unused:
        break;
    }
    statementVars->st_state = Unused;
    InformixApiDebug(3,function, "exit function");
    return 1;
}

/*
 *--------------------------------------------------------------
 *
 * InformixIntSetBindNum
 *
 * Creates the input descriptor for the specified number of items
 *
 * Results:
 *
 *--------------------------------------------------------------
*/
int
InformixIntSetBindNum(statementVars,items)
    StatementInfo   *statementVars;     /* Statement handle data */
    int             items;              /* Number of Items */
{
    static char function[] = "InformixIntSetBindNum";
    EXEC SQL BEGIN DECLARE SECTION;
    int  bind_size = items;
    char           *nm_ibind = statementVars->nm_ibind;
    EXEC SQL END DECLARE SECTION;

    InformixApiEnter(function);

    InformixIntSetConnection(statementVars->logonVars->nm_connection);

    if (items > statementVars->n_bound)
    {
        if (statementVars->n_bound > 0)
        {
            EXEC SQL DEALLOCATE DESCRIPTOR :nm_ibind;
            InformixApiSaveSqlca(statementVars->logonVars);
            statementVars->n_bound = 0;
            if (sqlca.sqlcode < 0)
            {
                InformixApiSqlCode(statementVars->logonVars);
                InformixApiExit(function);
                return 0;
            }
        }
        EXEC SQL ALLOCATE DESCRIPTOR :nm_ibind WITH MAX :bind_size;
        InformixApiSaveSqlca(statementVars->logonVars);
        if (sqlca.sqlcode < 0)
        {
            InformixApiSqlCode(statementVars->logonVars);
            InformixApiExit(function);
            return 0;
        }
        InformixApiDebug(2,"Descriptor %s allocated\n",statementVars->nm_ibind);
        statementVars->n_bound = items;
    }
    InformixApiExit(function);
    return 1;
}               

/*
 *--------------------------------------------------------------
 *
 * InformixIntClose
 *
 * Closes Cursor
 *
 * Results:
 *
 *--------------------------------------------------------------
*/
int
InformixIntClose(statementVars)
    StatementInfo   *statementVars;     /* Statement handle data */

{
    EXEC SQL BEGIN DECLARE SECTION;
    char           *nm_cursor = statementVars->nm_cursor;
    EXEC SQL END DECLARE SECTION;

    if (statementVars->st_state == Opened || 
        statementVars->st_state == Finished)
    {
        EXEC SQL CLOSE :nm_cursor;
        InformixApiSaveSqlca(statementVars->logonVars);
        if (sqlca.sqlcode < 0)
        {
            InformixApiSqlCode(statementVars->logonVars);
            return 0;
        }
        statementVars->st_state = Declared;
    }
    else
    {
        /*
            Nothing done if called in wrong state
        */
    }
    return 1;
}                   


/*
** InformixIntPreparse() -- based on dbd_preparse() in DBD::ODBC 0.15
**
** Edit string in situ (because the output will never be longer than the
** input) so that ? parameters are counted and :xx (x = digit) positional
** parameters are converted to ?.  Note that :abc notation is not
** converted; it causes problems with Informix's FROM dbase:table notation.
** Note that this version does not accept :x notation (where x is a digit)
** either, because it gets confused by DATETIME and INTERVAL literals!
** The code handles single-quoted literals and double-quoted delimited
** identifiers and ANSI SQL "--.*\n" comments and Informix "{.*}" comments.
** Note that it does nothing with "#.*\n" Perl/Shell comments.  Also note
** that it does not handle ODBC-style extensions.  The shorthand notation
** for these is identical to an Informix {} comment; longhand notation
** looks like "--*(details*)--" without the quotes.  Returns the number of
** input parameters.
*/

int 
InformixIntPreparse(statementVars,statement)
    StatementInfo   *statementVars;     /* Statement handle data */
    char            *statement;         /* Statement */
{                                            
    char            end_quote = '\0';
    char           *src;
    char           *dst;
    int             idx = 0;
    int             style = 0;
    int             laststyle = 0;
    char            ch;

    InformixApiDebug(4, "InformixIntPreparse in : <<%s>>\n", statement);
    src = statement;
    dst = statement;
    while ((ch = *src++) != '\0')
    {
        if (ch == end_quote)
            end_quote = '\0';
        else if (end_quote != '\0')
        {
            *dst++ = ch;
            continue;
        }                
        else if (ch == '\'' || ch == '\"')
            end_quote = ch;
        else if (ch == L_CURLY)
            end_quote = R_CURLY;
        else if (ch == '-' && *src == '-')
        {
            end_quote = '\n';
        }
        if (ch == '?')
        {
            /* X/Open standard   */
            *dst++ = '?';
            idx++;
            style = 3;
        }
        else
        {
            /* Perhaps ':=' PL/SQL construct or dbase:table in Informix */
            /* Or it could be :2 or :22 as part of a DATETIME/INTERVAL */
            *dst++ = ch;
            continue;
        }
        if (laststyle && style != laststyle)   {
            InformixApiSetError2(statementVars->logonVars,
                "Can't mix placeholder styles ");
            return 0;
        }
        laststyle = style;
    }
    if (end_quote != '\0')
    {
        switch (end_quote)
        {
        case '\'':
            InformixApiSetError2(statementVars->logonVars,
                "Incomplete single-quoted string");
            return 0;
        case '\"':
            InformixApiSetError2(statementVars->logonVars,
                "Incomplete double-quoted string (delimited identifier)");
            return 0;
        case R_CURLY:
            InformixApiSetError2(statementVars->logonVars, 
                "Incomplete bracketed {...} comment");
            return 0;
        case '\n':
            InformixApiSetError2(statementVars->logonVars,
                "Incomplete double-dash comment");
            break;
        default:
            break;
        }                          
    }
    *dst = '\0';
    InformixApiDebug(4, "InformixIntPreparse out: <<%s>>\n", statement);
    statementVars->n_input_params = idx;
    return(1);
}
          
int 
count_blobs(char *descname, int ncols)
{
    EXEC SQL BEGIN DECLARE SECTION;
    char           *nm_obind = descname;
    int colno;
    int coltype;
    EXEC SQL END DECLARE SECTION;
    int nblobs = 0;

    for (colno = 1; colno <= ncols; colno++)
    {
        EXEC SQL GET DESCRIPTOR :nm_obind VALUE :colno :coltype = TYPE;
        /* INFORMIX_sqlcode(imp_sth->dbh); */
        if (coltype == SQLBYTES || coltype == SQLTEXT)
        {
            nblobs++;
        }
    }
    return(nblobs);
}               
/*
** Fetch a single row of data.
**
** Note the use of 'varchar' variables.  Given the sample code:
**
** #include <stdio.h>
** int main(int argc, char **argv)
** {
**     EXEC SQL BEGIN DECLARE SECTION;
**     char    cc[30];
**     varchar vc[30];
**     EXEC SQL END DECLARE SECTION;
**     EXEC SQL WHENEVER ERROR STOP;
**     EXEC SQL DATABASE Apt;
**     EXEC SQL CREATE TEMP TABLE Test(Col01 CHAR(20), Col02 VARCHAR(20));
**     EXEC SQL INSERT INTO Test VALUES("ABCDEFGHIJ     ", "ABCDEFGHIJ     ");
**     EXEC SQL SELECT Col01, Col01 INTO :cc, :vc FROM Test;
**     printf("Col01: cc = <<%s>>\n", cc);             
**     printf("Col01: vc = <<%s>>\n", vc);
**     EXEC SQL SELECT Col02, Col02 INTO :cc, :vc FROM TestTable;
**     printf("Col02: cc = <<%s>>\n", cc);
**     printf("Col02: vc = <<%s>>\n", vc);
**     return(0);
** }
**
** The output looks like:
**      Col01: cc = <<ABCDEFGHIJ                   >>
**      Col01: vc = <<ABCDEFGHIJ          >>
**      Col02: cc = <<ABCDEFGHIJ                   >>
**      Col02: vc = <<ABCDEFGHIJ     >>
** Note that the data returned into 'cc' is blank padded to the length of
** the host variable, not the length of the database column, whereas 'vc'
** is blank-padded to the length of the database column for a CHAR column,
** and to the length of the inserted data in a VARCHAR column.
*/
int 
InformixIntFetch(statementVars,objc,objv,resList)

    StatementInfo   *statementVars;     /* Statement handle data */
    int             objc;               /* Number of output variables */
    Tcl_Obj         *CONST objv[];      /* List of var names */
    Tcl_Obj         *resList;           /* Result List */

{
    static char function[] = "InformixIntFetch";
    int            datatype;            /* result type */
    int            freemode;            /* free memory mode 
                                           0 do nothing
                                           1 free result 
                                           2 free lvarchr */
    EXEC SQL BEGIN DECLARE SECTION;
    char           *nm_cursor = statementVars->nm_cursor;
    char           *nm_obind = statementVars->nm_obind;
    varchar         coldata[256];
    long            coltype;
    long            collength;
    long            colind;
    char            colname[NAMESIZE];
    int             index;
    char           *result;
    long            length;
    loc_t           blob;
    lvarchar       *lvarchr;
    double          dblresult;
    long            longresult;
    int             intresult;
    float           floatresult;
    EXEC SQL END DECLARE SECTION;                    

    InformixApiEnter(function);
    
    /*
        Fetch row
    */

    EXEC SQL FETCH :nm_cursor USING SQL DESCRIPTOR :nm_obind;
    InformixApiSaveSqlca(statementVars->logonVars);
    if (sqlca.sqlcode != 0)
    {
        if (sqlca.sqlcode != SQLNOTFOUND)
        {
            InformixApiSqlCode(statementVars->logonVars);
            InformixApiDebug(1, "Exit %s -- fetch failed\n", function);
            InformixApiExit(function);
            return 0;
        }
        else
        {
            statementVars->st_state = Finished;
            InformixApiSqlCode(statementVars->logonVars);
            InformixApiDebug(1, "Exit %s -- SQLNOTFOUND\n", function);
            InformixApiExit(function);
            AppendResult(statementVars->logonVars->interp,
                "SQL: 100: no row found");
            return 0;
        }
    }
    statementVars->n_rows++;                   

    for (index = 1; index <= statementVars->n_columns; index++)
    {
        freemode=0;
        EXEC SQL GET DESCRIPTOR :nm_obind VALUE :index
                :coltype = TYPE, :collength = LENGTH,
                :colind = INDICATOR, :colname = NAME;
        InformixApiSqlCode(statementVars->logonVars);
        InformixApiDebug(6, "ColumnName %s \n", colname);
        InformixApiDebugl(6, "ColumnType %d \n", coltype);
        InformixApiDebugl(6, "ColumnLength %d \n", collength);
        InformixApiDebugl(6, "ColumnInicator %d \n", colind);

        if (colind != 0)
        {
            /* Data is null */
            result = coldata;
            length = 0;
            result[length] = '\0';
            datatype=R_CHAR;
        }
        else
        {
            switch (coltype)       { 
            case SQLDATE:
            case SQLDTIME:
            case SQLINTERVAL:
            case SQLDECIMAL:
            case SQLMONEY:       
                /* These types will always fit into a 256 character string */
                EXEC SQL GET DESCRIPTOR :nm_obind VALUE :index
                        :coldata = DATA;
                result = coldata;
                length = byleng(result, strlen(result));
                result[length] = '\0';
                InformixApiDebug(6,"Data= \"%s\"\n",result);
                datatype=R_CHAR;
                break;

            case SQLFLOAT:
                EXEC SQL GET DESCRIPTOR :nm_obind VALUE :index
                        :dblresult = DATA;
                datatype=R_DOUBLE;
                break;

            case SQLSMFLOAT:
                EXEC SQL GET DESCRIPTOR :nm_obind VALUE :index
                        :floatresult = DATA;
                dblresult= (double) floatresult ;
                datatype=R_DOUBLE;
                break;

            case SQLINT:
            case SQLSERIAL:
                EXEC SQL GET DESCRIPTOR :nm_obind VALUE :index
                        :longresult = DATA;
                datatype=R_LONG;
                break;

            case SQLSMINT:
                EXEC SQL GET DESCRIPTOR :nm_obind VALUE :index
                        :intresult = DATA;
                datatype=R_INT;
                break;

            case SQLVCHAR:
            case SQLNVCHAR:
                /* These types will always fit into a 256 character string */
                /* NB: VARCHAR strings always retain trailing blanks */
                EXEC SQL GET DESCRIPTOR :nm_obind VALUE :index
                        :coldata = DATA;
                result = coldata;
                length = strlen(result);
                InformixApiDebug(6,"Data= \"%s\"\n",result);
                datatype=R_CHAR;
                break;
                                
            case SQLCHAR:
            case SQLNCHAR:
                /**
                ** NB: CHAR strings have trailing blanks (which are added
                ** automatically by the database) removed by byleng() etc.
                */
                if (collength < 256)
                    result = coldata;
                else
                {
                    result = (char *) malloc(collength+1);
                    if (result == 0)
                    {
                        InformixApiSetError2(statementVars->logonVars,
                        "malloc failed in InformixIntFetch");
                        InformixApiExit(function);
                        return 0;
                    }
                    freemode=1;
                }
                EXEC SQL GET DESCRIPTOR :nm_obind VALUE :index
                        :result = DATA;
                length = strlen(result);       
                /* Conditionally chop trailing blanks */
                if(statementVars->logonVars->chop_blanks) 
                    length= byleng(result,length);
                result[length] = '\0';
                InformixApiDebug(6,"Data= \"%s\"\n",result);
                datatype=R_CHAR;
                break;

            case SQLTEXT:
            case SQLBYTES:
                blob_init(&blob);
                EXEC SQL GET DESCRIPTOR :nm_obind VALUE :index
                        :blob = DATA;                                                           result = blob.loc_buffer;
                /* Warning - this data is not null-terminated! */
                length = blob.loc_size;
                datatype=R_BINARY;
                freemode=0;
                InformixApiDebugl(6, "BlobLength %d \n", length);
                InformixApiDebug(6,"Data= \"%s\"\n","<Blob data>");
                break;

            case SQLLVARCHAR:
                ifx_var_flag(&lvarchr,1);
                EXEC SQL GET DESCRIPTOR :nm_obind VALUE :index
                        :lvarchr = DATA;
                length= ifx_var_getlen(&lvarchr);
                InformixApiDebugl(6, "lvaricharLength %d \n", length);
                result= (char *) ifx_var_getdata(&lvarchr);
                datatype=R_BINARY;
                freemode=2;
                break;

            case SQLUDTVAR:
                ifx_var_flag(&lvarchr,1);
                EXEC SQL GET DESCRIPTOR :nm_obind VALUE :index
                        :lvarchr = DATA;
                length= ifx_var_getlen(&lvarchr);
                InformixApiDebugl(6, "lvaricharLength %d \n", length);
                result= (char *) ifx_var_getdata(&lvarchr);
                datatype=R_BINARY;
                freemode=2;
                break;


            default:
                /*NOTREACHED*/               
                InformixApiSetError2(statementVars->logonVars,
                "unknown coltype in InformixIntFetch");
                length = 0;
                result = coldata;
                result[length] = '\0';
                break;
            }
            if (sqlca.sqlcode < 0)
            {
                InformixApiSqlCode(statementVars->logonVars);
                *result = '\0';
            }

        }
        /*
            Store results 
        */
        switch(datatype)
        {
        case R_BINARY: 
             if(objc==0) {
                    Tcl_ListObjAppendElement (statementVars->logonVars->interp,
                           resList, Tcl_NewByteArrayObj(result,length));
             } else {
                    Tcl_ObjSetVar2(statementVars->logonVars->interp,
                           objv[index-1],(Tcl_Obj *) NULL,
                           Tcl_NewByteArrayObj(result,length),
                           TCL_LEAVE_ERR_MSG);
             }
             break;

        case R_CHAR:
             if(objc==0) {
                    Tcl_ListObjAppendElement (statementVars->logonVars->interp,
                           resList, Tcl_NewStringObj(result,-1));
             } else {
                    Tcl_ObjSetVar2(statementVars->logonVars->interp,
                           objv[index-1],(Tcl_Obj *) NULL,
                           Tcl_NewStringObj(result,-1),
                           TCL_LEAVE_ERR_MSG);
             }
             break;

        case R_DOUBLE: 
             if(objc==0) {
                    Tcl_ListObjAppendElement (statementVars->logonVars->interp,
                           resList, Tcl_NewDoubleObj(dblresult));
             } else {
                    Tcl_ObjSetVar2(statementVars->logonVars->interp,
                           objv[index-1],(Tcl_Obj *) NULL,
                           Tcl_NewDoubleObj(dblresult),
                           TCL_LEAVE_ERR_MSG);
             }
             break;

        case R_LONG: 
             if(objc==0) {
                    Tcl_ListObjAppendElement (statementVars->logonVars->interp,
                           resList, Tcl_NewLongObj(longresult));
             } else {
                    Tcl_ObjSetVar2(statementVars->logonVars->interp,
                           objv[index-1],(Tcl_Obj *) NULL,
                           Tcl_NewLongObj(longresult),
                           TCL_LEAVE_ERR_MSG);
             }
             break;

        case R_INT: 
             if(objc==0) {
                    Tcl_ListObjAppendElement (statementVars->logonVars->interp,
                           resList, Tcl_NewIntObj(intresult));
             } else {
                    Tcl_ObjSetVar2(statementVars->logonVars->interp,
                           objv[index-1],(Tcl_Obj *) NULL,
                           Tcl_NewIntObj(intresult),
                           TCL_LEAVE_ERR_MSG);
             }
             break;
        }
        /*
          Clean up
        */
        if (freemode== 2) {
            ifx_var_dealloc(&lvarchr);
        } else if (freemode==1) {
            free(result);
        } else {
            /* do nothing */
        }
    }
    InformixApiExit(function);         
    return 1;
}

/*
 *--------------------------------------------------------------
 *
 * InformixIntOpen
 *
 * Opens Cursor Cursor
 *
 * Results:
 * 0 on error 1 on success
 *
 *--------------------------------------------------------------
*/
int
InformixIntOpen(statementVars)
    StatementInfo   *statementVars;     /* Statement handle data */

{
    static char function[] = "InformixIntOpen";
    EXEC SQL BEGIN DECLARE SECTION;
    char           *nm_cursor = statementVars->nm_cursor;
    char           *nm_ibind = statementVars->nm_ibind;
    EXEC SQL END DECLARE SECTION;

    InformixApiEnter(function);
    if (statementVars->st_state == Opened || 
        statementVars->st_state == Finished)
        InformixIntClose(statementVars);
    if (statementVars->n_bound > 0)
        EXEC SQL OPEN :nm_cursor USING SQL DESCRIPTOR :nm_ibind;
    else
        EXEC SQL OPEN :nm_cursor;
    InformixApiSaveSqlca(statementVars->logonVars);
    if (sqlca.sqlcode < 0)
    {                           
        InformixApiSqlCode(statementVars->logonVars);
        InformixApiExit(function);
        return 0;
    }
    statementVars->st_state = Opened;
    if (statementVars->logonVars->is_modeansi == True)
        statementVars->logonVars->is_txactive = True;
    statementVars->n_rows = 0;
    InformixApiExit(function);
    return 1;
}           
/*
 *--------------------------------------------------------------
 *
 * InformixIntBlobs
 *
 * Process Blobs
 *
 * Results:
 * 0 on error 1 on success
 *
 *--------------------------------------------------------------
 */

int
InformixIntBlobs(statementVars)
    StatementInfo   *statementVars;     /* Statement handle data */

{
    static char function[] = "InformixIntBlobs";
    EXEC SQL BEGIN DECLARE SECTION;
    char           *nm_obind = statementVars->nm_obind;
    loc_t          blob;
    int             colno;
    int coltype;
    EXEC SQL END DECLARE SECTION;
    int             n_columns = statementVars->n_columns;

    InformixApiEnter(function);
    statementVars->n_blobs = count_blobs(nm_obind, n_columns);
    if (statementVars->n_blobs == 0)
    {
        InformixApiExit(function);
        return 1;
    }


    /* Set blob location */
    if (blob_init(&blob) != 0)
    {
        InformixApiSetError2(statementVars->logonVars,
            "memory allocation error in InformixIntBlobs");
            InformixApiExit(function);
            return 0;
    }

    for (colno = 1; colno <= n_columns; colno++)
    {
        EXEC SQL GET DESCRIPTOR :nm_obind VALUE :colno :coltype = TYPE;
        InformixApiSqlCode(statementVars->logonVars);
        if (coltype == SQLBYTES || coltype == SQLTEXT)
        {
            /* Tell ESQL/C how to handle this blob */
            EXEC SQL SET DESCRIPTOR :nm_obind VALUE :colno DATA = :blob;
            InformixApiSqlCode(statementVars->logonVars);
        }
    }
    InformixApiExit(function);
    return 1;
}

/*
 *--------------------------------------------------------------
 *
 * InformixIntDeclare
 *
 * Declare Cursor for SELECT or EXECUTE PROCEDURE
 *
 * Results:
 * 0 on error 1 on success
 *
 *--------------------------------------------------------------
 */
int
InformixIntDeclare(statementVars)
    StatementInfo   *statementVars;     /* Statement handle data */

{
    static char function[] = "InformixIntDeclare";
    EXEC SQL BEGIN DECLARE SECTION;
    char           *nm_stmnt = statementVars->nm_stmnt;
    char           *nm_cursor = statementVars->nm_cursor;
    EXEC SQL END DECLARE SECTION;

    InformixApiEnter(function);
    InformixIntBlobs(statementVars);

    if (statementVars->logonVars->is_modeansi == True &&
                statementVars->logonVars->has_autocommit == True)
    {
        EXEC SQL DECLARE :nm_cursor CURSOR WITH HOLD FOR :nm_stmnt;
    }
    else
    {
        EXEC SQL DECLARE :nm_cursor CURSOR FOR :nm_stmnt;
    }
    InformixApiSaveSqlca(statementVars->logonVars);          
    if (sqlca.sqlcode < 0)
    {
        InformixApiSqlCode(statementVars->logonVars);
        return 0;
    }
    statementVars->st_state = Declared;
    InformixApiExit(function);
    return 1;
}        


/*
 *--------------------------------------------------------------
 *
 * InformixIntRelease
 *
 * Free up ressources used by statement or cursor
 *
 * Results:
 *
 *--------------------------------------------------------------
 */
int
InformixIntRelease(statementVars)
    StatementInfo   *statementVars;     /* Statement handle data */

{
    static char function[] = "InformixIntRelease";

    InformixApiEnter(function);
    InformixIntDelStatement(statementVars);
    InformixApiExit(function);
    return 1;
}

/*
 *--------------------------------------------------------------
 *
 * InformixIntExec
 *
 *
 * Results:
 *
 *--------------------------------------------------------------
 */
int
InformixIntExec(statementVars)
    StatementInfo   *statementVars;     /* Statement handle data */
{
    static char function[] = "InformixIntExec";
    EXEC SQL BEGIN DECLARE SECTION;
    char           *nm_stmnt = statementVars->nm_stmnt;
    char           *nm_ibind = statementVars->nm_ibind;
    EXEC SQL END DECLARE SECTION;
    LogonInfo      *logonVars = statementVars->logonVars;
    int rc = 1;
    Boolean exec_stmt = True;

    InformixApiEnter(function);

    if (statementVars->st_type == SQ_BEGWORK)
    {
        /* BEGIN WORK in a logged non-ANSI database with AutoCommit Off */
        /* will fail because we're already in a transaction. */
        /* Pretend it succeeded. */
        if (logonVars->is_loggeddb == True && logonVars->is_modeansi == False)
        {
            if (logonVars->has_autocommit == False)
            {
                exec_stmt = False;
                sqlca.sqlcode = 0;                
            }
        }
    }

    if (exec_stmt == True)
    {
        if (statementVars->n_bound > 0)
        {
            EXEC SQL EXECUTE :nm_stmnt USING SQL DESCRIPTOR :nm_ibind;
        }
        else
        {
            EXEC SQL EXECUTE :nm_stmnt;
        }
    }

    InformixApiSaveSqlca(logonVars);
    if (sqlca.sqlcode < 0)
    {
        InformixApiSqlCode(logonVars);
        InformixApiExit(function);
        return 0;                
    }

    /**
    ** Here we need to analyse what was done...
    ** BEGIN WORK, COMMIT WORK, ROLLBACK WORK are important.
    ** So are DATABASE, CLOSE DATABASE, CREATE DATABASE.
    ** For SE, we could use START DATABASE or ROLLFORWARD DATABASE.
    ** Note that although it is unlikely to happen with Perl, the DATABASE
    ** operations other than CLOSE DATABASE can have a '?' place of the
    ** database name, so the same statement could be executed several times
    ** with different names, and the name is then available in nm_ibind.
    ** On the other hand, if it is not in nm_ibind, it has to be extracted
    ** from the statement string itself.
    */
    statementVars->n_rows = sqlca.sqlerrd[2];
    switch (statementVars->st_type)
    {
    case SQ_BEGWORK:
        InformixApiDebug(3, "%s: BEGIN WORK\n", function);
        logonVars->is_txactive = True;
        /* Even BEGIN WORK has to be committed if AutoCommit is On */
        if (logonVars->has_autocommit == True)           
            rc = InformixIntCommit(logonVars);
        break;
    case SQ_COMMIT:
        InformixApiDebug(3, "%s: COMMIT WORK\n", function);
        logonVars->is_txactive = False;
        /* In a logged database with AutoCommit Off, do BEGIN WORK */
        if (logonVars->is_modeansi == False && 
            logonVars->has_autocommit == False)
            rc = InformixIntBegin(logonVars);
        break;
    case SQ_ROLLBACK:
        InformixApiDebug(3, "%s: ROLLBACK WORK\n", function);
        logonVars->is_txactive = False;
        /* In a logged database with AutoCommit Off, do BEGIN WORK */
        if (logonVars->is_modeansi == False && 
            logonVars->has_autocommit == False)
            rc = InformixIntBegin(logonVars);
        break;
    case SQ_DATABASE:
        logonVars->is_txactive = False;
        InformixIntSetDbType(logonVars);
        /* Analyse new database name and record it */           break;
    case SQ_CREADB:
        logonVars->is_txactive = False;
        InformixIntSetDbType(logonVars);
        /* Analyse new database name and record it */
        break;
    case SQ_STARTDB:
        logonVars->is_txactive = False;
        InformixIntSetDbType(logonVars);
        /* Analyse new database name and record it */
        break;
    case SQ_RFORWARD:
        logonVars->is_txactive = False;
        InformixIntSetDbType(logonVars);
        /* Analyse new database name and record it */
        break;
    case SQ_CLSDB:
        logonVars->is_txactive = False;
        /* Record that no database is open */
        break;
    default:
        if (logonVars->is_modeansi)
            logonVars->is_txactive = True;                
        /* COMMIT WORK for MODE ANSI databases when AutoCommit is On */
        if (logonVars->is_modeansi == True && 
            logonVars->has_autocommit == True)
            rc = InformixIntCommit(logonVars);
        break;
    }

    InformixApiExit(function);
    return rc;
}           

/*
 *--------------------------------------------------------------
 *
 * InformixIntSetDbType
 *
 *
 * Results:
 *
 *--------------------------------------------------------------
 */
void
InformixIntSetDbType(logonVars)
    LogonInfo       *logonVars;         /* Logon handle data */
{
    logonVars->is_onlinedb = (sqlca.sqlwarn.sqlwarn3 == 'W');
    logonVars->is_modeansi = (sqlca.sqlwarn.sqlwarn2 == 'W');
    logonVars->is_loggeddb = (sqlca.sqlwarn.sqlwarn1 == 'W');
}  
/*
 *--------------------------------------------------------------
 *
 * InformixIntBind
 * Bind Parameters
 *
 * Results:
 *
 *--------------------------------------------------------------
 */
int
InformixIntBind(statementVars, count, obj)
    StatementInfo       *statementVars;     /* Statement handle data */
    int                 count;              /* Pos of Param to bind */
    Tcl_Obj             *obj;               /* TCL-Object */
{
    int                 rc=1;
    static char         function[] = "InformixIntBind";
    int                 len;
    char                *bbuffer;

    EXEC SQL BEGIN DECLARE SECTION;
    char           *nm_ibind = statementVars->nm_ibind;
    char *string;
    int     type;
    int     length;
    int index = count+1;
    loc_t blob;
    EXEC SQL END DECLARE SECTION;
                                   
    InformixApiEnter(function);

    InformixApiDebug(2,"Descriptor %s\n",statementVars->nm_ibind);
    EXEC SQL GET DESCRIPTOR :nm_ibind VALUE :index :type = TYPE;
    InformixApiSaveSqlca(statementVars->logonVars);

    if (sqlca.sqlcode < 0)
    {
        InformixApiSqlCode(statementVars->logonVars);
        InformixApiExit(function);
        return 0;
    }
    InformixApiDebugl(2,"Data Type %d\n",type);
/*
    Blob data
*/

    if (type == SQLBYTES || type == SQLTEXT)
    {
        /* One day, this will accept SQ_UPDATE and SQ_UPDALL */
        /* There are no plans to support SQ_UPDCURR */
        bbuffer = Tcl_GetByteArrayFromObj(obj,&len);
        InformixApiDebugl(2, "BLOB Data len=%d \n", len);

        if (len == 0 ) {
             InformixApiDebug(2, "%s -- null\n", function);
             type = SQLCHAR;
             EXEC SQL SET DESCRIPTOR :nm_ibind VALUE :index
                        TYPE = :type, LENGTH = 0, INDICATOR = -1;
        } else {
             blob_init(&blob);
             blob.loc_buffer = bbuffer;
             blob.loc_bufsize = len;
             blob.loc_size = len;
             EXEC SQL SET DESCRIPTOR :nm_ibind VALUE :index DATA = :blob;
        }                            
    }
    else
    {
        type = SQLCHAR;
        string = Tcl_GetStringFromObj(obj,&len); 
        InformixApiDebugl(2, "String data len=%d\n", len);
        if (len == 0 ) {
             InformixApiDebug(2, "%s -- null\n", function);
             type = SQLCHAR;
             EXEC SQL SET DESCRIPTOR :nm_ibind VALUE :index
                        TYPE = :type, LENGTH = 0, INDICATOR = -1;
        } else {
             length=len+1;
             if (length == 1)
             {
                 /*
                   Even if you insert "" as a literal into a VARCHAR(), you 
                   get ** a blank returned.  If you manage to insert a zero 
                   length string via a variable into a VARCHAR, then you get 
                   a NULL output string.  This is arguably a bug, but oh well.
                 */
                 string = " ";
                 length = 2;
             }
             InformixApiDebug(6, "data=%s\n", string);
             EXEC SQL SET DESCRIPTOR :nm_ibind VALUE :index
                        TYPE = :type, LENGTH = :length, DATA = :string;
        }
    }                                         
    InformixApiSqlCode(statementVars->logonVars);
    if (sqlca.sqlcode < 0)
    {
        rc = 0;
    }
    InformixApiDebugl(6, "ret code=%d\n", rc);
    InformixApiExit(function);
    return(rc);
}            
/*
 *--------------------------------------------------------------
 *
 * InformixIntGetAttrib
 * Get Column Attributes
 *
 * Results:
 *
 *--------------------------------------------------------------
 */
int
InformixIntGetAttrib(statementVars, ind, itype, retlong, retstring)
    StatementInfo       *statementVars;     /* Statement handle data */
    int                 itype;              /* Type of information */
    long                *retlong;           /* Ret Val Integer */
    char                *retstring;         /* Ret Val String */
    int                 ind;                /* Index of Col */

{
    static              char function[]= "InformixIntGetAttrib";

    EXEC SQL BEGIN DECLARE SECTION;
    char           *nm_obind = statementVars->nm_obind;
    long            coltype;
    long            collength;
    long            colnull;
    long            colscale;
    long            colprecision;
    char            colname[NAMESIZE];
    int             i;
    EXEC SQL END DECLARE SECTION;       

    InformixApiEnter(function);
    i=ind;
    switch(itype) {

    /*
        Get Column Names
    */
    case 0:             
        EXEC SQL GET DESCRIPTOR :nm_obind VALUE :i
                :colname = NAME;
            colname[byleng(colname, strlen(colname))] = '\0';
            strcpy(retstring,colname);
            break;
    /*
        Get Nullable Info
    */

    case 1:
        EXEC SQL GET DESCRIPTOR :nm_obind VALUE :i
                :colnull = NULLABLE;  
        *retlong= colnull;
        break;

    /*
        Get Datatype
    */
    case 2:
        EXEC SQL GET DESCRIPTOR :nm_obind VALUE :i
                :coltype = TYPE, :collength = LENGTH;    
        switch(coltype) {
        case SQLCHAR:    strcpy(retstring,"SQLCHAR");
                break;
        case SQLSMINT:   strcpy(retstring,"SQLSMINT");
                break;
        case SQLINT:     strcpy(retstring,"SQLINT");
                break;
        case SQLFLOAT:   strcpy(retstring,"SQLFLOAT");
                break;
        case SQLSMFLOAT: strcpy(retstring,"SQLSMFLOAT");
                break;
        case SQLDECIMAL: strcpy(retstring,"SQLDECIMAL");
                break;
        case SQLSERIAL: strcpy(retstring,"SQLSERIAL");
                break;
        case SQLDATE: strcpy(retstring,"SQLDATE");
                break;
        case SQLMONEY: strcpy(retstring,"SQLMONEY");
                break;
        case SQLNULL: strcpy(retstring,"SQLNULL");
                break;
        case SQLDTIME: strcpy(retstring,"SQLDTIME");
                break;
        case SQLBYTES: strcpy(retstring,"SQLBYTES");
                break;
        case SQLTEXT: strcpy(retstring,"SQLTEXT");
                break;
        case SQLVCHAR: strcpy(retstring,"SQLVCHAR");
                break;
        case SQLINTERVAL: strcpy(retstring,"SQLINTERVAL");
                break;
        case SQLNCHAR: strcpy(retstring,"SQLNCHAR");
                break;
        case SQLNVCHAR: strcpy(retstring,"SQLNVCHAR");
                break;
        }
    /*
        Length
    */
    case 3:
        EXEC SQL GET DESCRIPTOR :nm_obind VALUE :i
                :collength = LENGTH;   
        *retlong= collength;
        break;

    /*
        Scale
    */
    case 4:
        EXEC SQL GET DESCRIPTOR :nm_obind VALUE :i
                :colscale = SCALE;   
        *retlong= colscale;
        break;

    /*
        Precision
    */
    case 5:
        EXEC SQL GET DESCRIPTOR :nm_obind VALUE :i
                :colprecision = PRECISION;   
        *retlong= colprecision;
        break;
    }

    if (sqlca.sqlcode < 0)
    {
        InformixApiSqlCode(statementVars->logonVars);
        InformixApiExit(function);
        return 0;
    }
    InformixApiExit(function);
    return(1);
}

/*
 *--------------------------------------------------------------
 *
 * InformixIntLogoff
 * Logs Off Database
 *
 * Results:
 * 0 on error
 * 1 on success
 *
 *--------------------------------------------------------------
 */
int InformixIntLogoff (logonVars)
    LogonInfo       *logonVars;         /* Logon handle data */  
{
    InformixIntDisconnect(logonVars->nm_connection);

    if(sqlca.sqlcode < 0 ) {
        InformixApiSetError(logonVars->interp, sqlca.sqlcode);
        return(0);
    }
    return(1);
}
                
/*
 *--------------------------------------------------------------
 *
 * InformixIntLogon
 * Logs on to Database
 *
 * Results:
 * 0 on error
 * 1 on success
 *
 *--------------------------------------------------------------
 */
int InformixIntLogon (logonVars, name,user,pass)
    LogonInfo       *logonVars;         /* Logon handle data */  
    char            *name;              /* DB Name */
    char            *user;              /* User Name */
    char            *pass;              /* Password */

{
    if (user != 0 && *user == '\0')
        user = 0;
    if (pass != 0 && *pass == '\0')
        pass = 0;
    /* 6.00 and later versions of Informix-ESQL/C support CONNECT */
    InformixIntConnect(logonVars->nm_connection, name, user, pass);
    /*
        Failure ?
    */
    if(sqlca.sqlcode < 0 ) {
        InformixApiSetError(logonVars->interp, sqlca.sqlcode);
        return 0;
    }                        
    return 1;
}

/* Internal implementation of Dosql */

int 
InformixIntDosql(logonVars,cmd)
    LogonInfo       *logonVars;     /* logon handle */
    char            *cmd;           /* SQL Statement */
{
    EXEC SQL BEGIN DECLARE SECTION;
    char           *statement = cmd;
    EXEC SQL END DECLARE SECTION;
    int             rc = 1;

    EXEC SQL EXECUTE IMMEDIATE :statement;
    if (sqlca.sqlcode < 0) {
        InformixApiSqlCode(logonVars);
        rc=0;
    }
    return rc;
}         

/*
** Use CONNECT to initiate database connection
**
** If both user and password are provided, then the USER clause is used.
** If no database is specified, a default connection will be made.
**
** Note that CONNECT statements (and DISCONNECT and SET CONNECTION)
** cannot be prepared.
*/

int 
InformixIntConnect(char *connection, char *dbase, char *user, char *pass)
{
	EXEC SQL BEGIN DECLARE SECTION;
	char           *dbconn;
	char           *dbname;
	char           *dbpass;
	char           *dbuser;
	EXEC SQL END DECLARE SECTION;
	Boolean         conn_ok = False;

	if (dbase == (char *)0 || *dbase == '\0')
	{
		/* Not frequently used, but valid */
		/* Reset connection name to empty string, and connect to default */
		/* Typically used to create database on default server */
		/* Nasty interface, overwriting connection name! */
		*connection = '\0';
		InformixApiDebug(1, "CONNECT TO DEFAULT%s\n", connection);
		EXEC SQL CONNECT TO DEFAULT
			WITH CONCURRENT TRANSACTION;
	}
	else if (user != (char *)0 && pass != (char *)0)
	{
		dbconn = connection;
		dbname = dbase;
		dbpass = pass;
		dbuser = user;
		InformixApiDebug(1, "CONNECT with user info (%s)\n", connection);
		EXEC SQL CONNECT TO :dbname AS :dbconn
			USER :dbuser USING :dbpass
			WITH CONCURRENT TRANSACTION;
	}
	else
	{
		dbconn = connection;
		dbname = dbase;
		InformixApiDebug(1, "CONNECT - no user info (%s)\n", connection);
		EXEC SQL CONNECT TO :dbname AS :dbconn
			WITH CONCURRENT TRANSACTION;
	}
	if (sqlca.sqlcode == 0)
		conn_ok = True;
	return(conn_ok);
}

/* Basic interface to DISCONNECT */
static void do_disconnect(char *connection)
{
	EXEC SQL BEGIN DECLARE SECTION;
	char           *dbconn = connection;
	EXEC SQL END DECLARE SECTION;

	if (*connection != '\0')
	{
		InformixApiDebug(1, "DISCONNECT (%s)\n", connection);
		EXEC SQL DISCONNECT :dbconn;
	}
	else
	{
		InformixApiDebug(1, "DISCONNECT DEFAULT%s\n", connection);
		EXEC SQL DISCONNECT DEFAULT;
	}
}

/* External interface to disconnect which handles various oddities */
void 
InformixIntDisconnect(char *connection)
{
	do_disconnect(connection);
	if (sqlca.sqlcode == -1800)
	{
		/*
		** -1800: Invalid transaction state
		** One problem was discovered by Nathan Neulinger (nneul@umr.edu).
		** This can occur if the application is talking to a 5.0x engine.
		** The solution seems to be to do CLOSE DATABASE, which also closes
		** the connection (you get error -1803 if you try to DISCONNECT
		** after doing CLOSE DATABASE).  Ensure that the correct connection
		** is current before closing the database.  Bugs B64926 and B42204
		** are the source of the trouble.
		**
		** A second version of the problem found with 7.x databases where a
		** transaction has been started and not completed.  Trying CLOSE
		** DATABASE fails with -759 (Cannot use database commands in an
		** explicit database connection).  It requires a ROLLBACK WORK
		** instead, followed by a DISCONNECT.  If you are connected to a
		** 5.0x engine, then the ROLLBACK WORK succeeds, but the DISCONNECT
		** fails a second time, but the CLOSE DATABASE ruse then works.
		**
		** If both these attempts fail, then the disconnect operation gives
		** up, letting the database engine clean up.  The engine will do a
		** rollback when it notes the absence of the application.
		*/
		InformixApiDebug(1, "DISCONNECT **FAILED: -1800 ** <<%s>>\n", connection);
		InformixIntSetConnection(connection);
		InformixApiDebug(1, "Try ROLLBACK WORK <<%s>>\n", connection);
		EXEC SQL ROLLBACK WORK;
		if (sqlca.sqlcode == 0)
		{
			InformixApiDebug(1, "ROLLBACK WORK worked <<%s>>\n", connection);
			do_disconnect(connection);
		}
		if (sqlca.sqlcode < 0)
		{
			InformixApiDebugl(1, "DISCONNECT ** FAILED AGAIN (%ld) **\n",
						   sqlca.sqlcode);
			InformixApiDebug(1, "Try CLOSE DATABASE <<%s>>\n", connection);
			EXEC SQL CLOSE DATABASE;
			if (*connection == '\0')
			{
				InformixApiDebug(1, "Retry DISCONNECT DEFAULT%s\n", connection);
				EXEC SQL DISCONNECT DEFAULT;
			}
		}
	}
	InformixApiDebugl(1, "DISCONNECT -- STATUS %ld\n", sqlca.sqlcode);
}

/* Ensure that the correct connection is current -- a no-op in version 5.0x */
void InformixIntSetConnection(char *conn)
{
	EXEC SQL BEGIN DECLARE SECTION;
	char           *nm_connection = conn;
	EXEC SQL END DECLARE SECTION;

	if (nm_connection)
	{
		InformixApiDebug(1, "SET CONNECTION %s\n", nm_connection);
		EXEC SQL SET CONNECTION :nm_connection;
	}
	else
	{
		InformixApiDebug(1, "SET CONNECTION DEFAULT%s\n", "");
		EXEC SQL SET CONNECTION DEFAULT;
	}
}
/*
** Initialize a Blob data structure ready for use.
** Returns: 0 => OK, non-zero => fail
*/
int blob_init(Blob * blob)
{
    static char     function[]= "blob_init";
    int rc;

    InformixApiEnter(function);
	rc=0;
    blob->loc_status = 0;
    blob->loc_type = SQLTEXT;
    blob->loc_xfercount = 0;

    InformixApiExit(function);
    /* Use memory only */
    blob->loc_loctype = LOCMEMORY;
    blob->loc_size = 0;
    blob->loc_bufsize = -1;
    blob->loc_buffer = (char *)0;
    blob->loc_indicator = 0;
    blob->loc_oflags = 0;
    blob->loc_mflags = 0;  
    InformixApiExit(function);
    return(rc);
}                  
